!************************************************************************!
!Resolution de l'equation de Poisson en utilisant Jacobi
!     sur le domaine [0,1]x[0,1] par une methode aux differences finies,
!     avec comme solveur Jacobi.
!     Delta  u =  f(x,y) = 2*(x*x-x+y*y-y)
!     u sur les bords vaut 0
!     La solution exacte est u =  x*y*(x-1)*(y-1)
!
!     La valeur de u est donnee par la formule :
!     u(i,j)(n+1) = 0.25*(u(i+1,j)(n)+u(i-1,j)(n) +
!                         u(i,j+1)(n)+u(i,j-1)(n) - h*h*f(i,j))
!
!     Dans cette version, on se donne le nombre de points interieurs
!     total suivant x (ntx) et le nombre de points inferieurs total
!     suivant y (nty).
!     ntx = nty
!     hx represente le pas suivant x , hy le pas suivant y, comme
!     ntx = nty, hx = hy = h = 1/(ntx+1)
!
!     Pour chaque processeur :
!     1) decomposer le domaine
!     2) connaitre ses 4 voisins
!     3) echanger les points aux interfaces
!     4) calculer
!
!     Auteur           : Isabelle DUPAYS (CNRS/IDRIS - France)
!                        <Isabelle.Dupays@idris.fr>
!     Cree le          : Fri nov 15 10:32:45 1996
!***********************************************************************!
!
      PROGRAM Poisson
!
      USE MPI
!
      USE numerics
!
      IMPLICIT   NONE
!
      INCLUDE     "interface.h"
!
!---Declaration  des  variables
!
!Solution  u  a  l'iteration  n  et  n+1
      REAL(rp),  DIMENSION(:,:,:),  ALLOCATABLE  ::  u
!
!Solution exacte
      REAL(rp),  DIMENSION(:,:),    ALLOCATABLE  ::  u_exact
!
!Second membre
      REAL(rp),  DIMENSION(:,:),    ALLOCATABLE  ::  f
!
!Nombre  total  de points  interieurs suivant x  et  y
      INTEGER  ::  ntx, nty
!
!Borne  inferieure  et  superieure  en  x
      INTEGER  ::  sx, ex
!
!Borne  inferieure  et  superieure  en  y
      INTEGER  ::  sy, ey
!
!Pas
      REAL(rp)  ::  h
!
!Tableau contenant les voisins du domaine courant
      INTEGER,  PARAMETER                  ::  NB_VOISINS  =  4
!
      INTEGER,  DIMENSION(NB_VOISINS)      ::  voisin
      INTEGER,  PARAMETER                  ::  N=1,  E=2,   S=3,  W=4
!
!Nombre  iterations  en  temps
      INTEGER   ::  it
!
!Nombre  iterations  maximum  en  temps
      INTEGER, PARAMETER                   ::  it_max=10000
!
!Convergence
      REAL(rp)  ::  diffnorm, dwork
!
!Mesure  du  temps
      REAL(rp)  ::  t1, t2
!
!Compteur
      INTEGER   ::  j
!
!Test  de  convergence
      LOGICAL   ::  convergence
!
!Argument  muet  de  la  fonction  F90  EPSILON
      REAL(rp)  ::  X
      REAL(rp)  ::  machep
!
!Constantes  MPI
!
      INTEGER   ::  comm2d, type_ligne
      INTEGER,  DIMENSION(2)  ::  dims
      LOGICAL,  DIMENSION(2)  ::  periods
!
     !***********************************************************************!
!
!Initialisation  de  MPI
      CALL  MPI_INIT(ierr)
!
!Savoir  quel  processeur  je  suis
      CALL  MPI_COMM_RANK(  MPI_COMM_WORLD,  myrank,  ierr)
!
!Connaitre le nombre total de processeurs
      CALL  MPI_COMM_SIZE(  MPI_COMM_WORLD,  nbproc,  ierr)
!
      WRITE (6, 1010) 'Je  suis  le  processeur',  myrank  ,'  sur  ',  nbproc,  '  processeurs'
 1010 FORMAT ( A, I3, A, I3, A )
!
!Pour le processeur  0, connaitre  le  nombre  de  points  interieurs  total  en  x
      machep = MAX( 100.0_rp * EPSILON(X), 1E-7 )
!
      IF  (myrank  ==  0)  THEN
          read  (5,*) ntx
          write (6,1020) ' machep = ', machep
 1020 FORMAT ( A, E12.4 )
      END IF
!
!Type de donnees
      IF ( rp == dp ) THEN
         ITYPE_REAL = MPI_DOUBLE_PRECISION
      ELSE
         ITYPE_REAL = MPI_REAL
      END IF
!
!Envoi de la valeur de ntx a tous les autres processeurs
      CALL  MPI_BCAST(  ntx,  1,  MPI_INTEGER,   0,  MPI_COMM_WORLD,  ierr)
!
!Determiner  le  nombre  de  points  interieurs  total  en  y  nty  =  ntx
      nty  =  ntx
!
!Connaitre  le  nombre  de  processeurs  selon  x  et  le  nombre  de  processeurs
!selon  y  en  fonction  du  nombre  total  de  processeurs
      dims  =  0
      CALL  MPI_DIMS_CREATE(  nbproc,  2,  dims,  ierr)
!
!Creation  de  la  grille  de  processeurs  2D  sans  periodicite
      periods  =  .FALSE.
!
      CALL  MPI_CART_CREATE( MPI_COMM_WORLD,  2,  dims,  periods,  .TRUE., &
                             comm2d,  ierr)
!
!Recherche  de  ses  4  voisins  pour  chaque  processeur
      CALL  voisinage(  comm2d,  voisin)
      WRITE (6, 1030) 'Processeur  ',  myrank,  ':',  voisin
 1030 FORMAT ( A, I3, A, 4I3)
!
!Determiner  les  indices  de  chaque  sous-domaine
      CALL domaine(  comm2d,  ntx,  sx,  ex,  sy,  ey)
!
!Allocation  dynamique  des  tableaux  u,u_exact,f
      ALLOCATE(  u      (sx-1:ex+1,  sy-1:ey+1, 0:1), &
                 f      (sx-1:ex+1,  sy-1:ey+1), &
                 u_exact(sx-1:ex+1,  sy-1:ey+1), STAT = ierr )
      IF ( ierr /= 0 ) THEN
         WRITE (6,1010) 'rang = ', myrank
         WRITE (6,1010) 'ierr = ', ierr
         WRITE (6,1010) 'Problemne allocation'
         CALL MPI_ABORT ( MPI_COMM_WORLD, 1, ierr )
         STOP
      END IF
!
!Creation  du  type  type_ligne  pour  echanger  les  points  a  l'ouest  et  a  l'est
      CALL  MPI_TYPE_VECTOR(  ey-sy+1,  1,  ex-sx+3,  ITYPE_REAL,  type_ligne,  ierr)
      CALL  MPI_TYPE_COMMIT(  type_ligne,  ierr)
!
!Initialisation  du  second  membre  et  de  la  solution  u  sur  les  bords
      CALL  initialisation(  u,  f,  u_exact,  ntx,  h,  sx,  sy)
!
!Barriere
      CALL  MPI_BARRIER(  comm2d,  ierr)
!
      IF ( myrank == 0 ) THEN
         OPEN (UNIT=26, FILE='fort.26', STATUS='UNKNOWN', FORM='FORMATTED')
      END IF
!
!Schema  iteratif  en  temps
      it  =  0
      convergence  =  .FALSE.
!
!Mesure  du  temps  en  seconde  dans  la  boucle  en  temps
      t1  =  MPI_WTIME()
!
      DO  WHILE  ( (.NOT. convergence)  .AND.  (it  <  it_max) )
!
         it  =  it  +  1
         j = iold
         iold = inew
         inew = j
!
!Echange  des  points  aux  interfaces  pour  u  a  l'iteration  n
         CALL communication(  u,  sx,  sy,  type_ligne,  voisin,  comm2d)
!
!Calcul  de  u  a  l'iteration  n+1
         CALL  calcul(  u,  f,  sx,  sy,  h)
!
!Calcul  de  l'erreur  sur  le sous-domaine
         dwork  =  erreur(  u,  sx,  sy)
!
!Calcul  de  l'erreur  sur  tous  les  sous-domaines
         CALL  MPI_ALLREDUCE( dwork,  diffnorm,  1,  ITYPE_REAL,  &
                              MPI_MAX,  comm2d,  ierr )
!
!Arret  du  programe  si  on  a  atteint  la  precision  machine  obtenu
!par  la  fonction  F90  EPSILON
         convergence  =  ( diffnorm  <  machep )
!
!Affichage  pour  le  processeur  0  de  la  difference
         IF  ( (myrank == 0) .AND. (MOD(it,100)  ==  0) )  THEN
            WRITE (26,1080) it,  'La  difference est  ',  diffnorm
            CALL FLUSH (26)
 1080       FORMAT (1x, I5, 2x, A, E12.5)
         END  IF
!
      END  DO
!
!Mesure  du  temps  a  la  sortie  de  la  boucle
      t2  =  MPI_WTIME()
!
!Affichage  du  temps  de  convergence  par  le  processeur  0
      IF  ( myrank  ==  0 )  THEN
         WRITE (6,1040)  'Iteration : ',  it,  ' la difference est  ',  diffnorm
         WRITE (6,1050)  'Temps de calcul : ', t2 - t1
         WRITE (26,1060) 'Solution  exacte  u_exact  ',  'Solution  calculee  u'
 1040 FORMAT ( A, I6, A, E12.4 )
 1050 FORMAT ( A, E12.4 )
 1060 FORMAT ( 2A )
         DO  j  =  sy,  ey
             WRITE (26,1070) u_exact(1,j), u(1,j,inew), ABS( u_exact(1,j) - u(1,j,inew) )
 1070 FORMAT  ('u_exact  =   ', E17.10,  '  u  =   ', E17.10, &
               '  diff  =  ', E17.10)
         END  DO
      END  IF
!
      IF ( myrank == 0 ) CLOSE (26)
!
!Desallocation  des  tableaux  u,  u_exact  et  f
      DEALLOCATE(  u,  u_exact,  f)
!
!Liberation  du  type  type_ligne  et  du  communicateur  comm2d
      CALL  MPI_TYPE_FREE( type_ligne,  ierr)
      CALL  MPI_COMM_FREE( comm2d,  ierr)
!
!Desactivation  de  MPI
      CALL  MPI_FINALIZE(ierr)
!
      END  PROGRAM  Poisson

